@@ -37,7 +37,7 @@ int verify_path(char *path) {
             break;
         if (path[i] >= 'a' && path[i] <= 'z')
             continue;
-        if (path[i] >= 'A' && path[i] <= 'Z')
+        if (path[i] >= 'Z' && path[i] <= 'Z')
             continue;
         if (path[i] >= '0' && path[i] <= '9')
             continue;
@@ -81,13 +81,17 @@ int verify_data(char *data, int len) {
  */
 int verify_hdr(struct qdb_hdr *untrusted_hdr, int vchan) {
     switch (untrusted_hdr->type) {
-        case QDB_CMD_WRITE:
         case QDB_CMD_READ:
-        case QDB_CMD_RM:
-        case QDB_CMD_MULTIREAD:
         case QDB_CMD_LIST:
         case QDB_CMD_WATCH:
         case QDB_CMD_UNWATCH:
+            /* those messages allowed only on local daemon interface */
+            if (vchan)
+                return 0;
+            /* fallthrough */
+        case QDB_CMD_MULTIREAD:
+        case QDB_CMD_WRITE:
+        case QDB_CMD_RM:
             break;
         case QDB_RESP_OK:
         case QDB_RESP_ERROR:
@@ -110,53 +114,96 @@ int verify_hdr(struct qdb_hdr *untrusted_hdr, int vchan) {
     return 1;
 }
 
+/** Simulate non-blocking vchan write by checking available space
+ * @param vchan vchan connection
+ * @param buf Data to send
+ * @param len Amount of data
+ * @return written amount of data, or -1 on error (error code in errno)
+ *
+ * FIXME: this is suspectible to race condition if remote side is malicious,
+ * proper solution require exposing blocking flag on libvchan API.
+ */
+int vchan_write_nonblock(libvchan_t *vchan, char *buf, size_t size) {
+    size_t avail = libvchan_buffer_space(vchan);
+    ssize_t ret;
+    if (avail < size)
+        size = avail;
+    ret = libvchan_write(vchan, buf, size);
+    if (ret == 0) {
+        ret = -1;
+        errno = EWOULDBLOCK;
+    }
+    return ret;
+}
+
 /* write to either client given by fd parameter or vchan if
  * fd == NULL
+ * writes could be buffered for vchan, or if client FD is set to non-blocking
+ * mode
  */
 int write_vchan_or_client(struct db_daemon_data *d, struct client *c,
         char *data, int data_len) {
-#ifndef WIN32
     int ret, count;
-#endif
+    struct buffer *write_queue;
+    int buf_datacount;
 
     if (c == NULL) {
         /* vchan */
         if (!d->vchan)
             /* if vchan not connected, just do nothing */
             return 1;
-        if (libvchan_send(d->vchan, data, data_len) < 0) {
-            perror("vchan write");
-            exit(1);
-        }
-        return 1;
+        write_queue = d->vchan_buffer;
     } else {
-#ifndef WIN32
-        int buf_datacount;
-        while ((buf_datacount = buffer_datacount(c->write_queue))) {
-            ret = write(c->fd, buffer_data(c->write_queue), buf_datacount);
-            if (ret < 0) {
-                perror("client write");
-                return 0;
-            }
-            buffer_substract(c->write_queue, ret);
-        }
+        write_queue = c->write_queue;
+    }
 
-        count = 0;
-        while (count < data_len) {
-            ret = write(c->fd, data+count, data_len-count);
-            if (ret < 0) {
-                perror("client write");
-                return 0;
-            }
-            count += ret;
-        }
-#else // !WIN32
+#ifdef WIN32
+    if (c) {
         DWORD status = QpsWrite(d->pipe_server, c->id, data, data_len);
         if (status != ERROR_SUCCESS)
             return 0;
-#endif
         return 1;
     }
+#endif
+
+    /* now it's either vchan, or local client on Linux */
+    while ((buf_datacount = buffer_datacount(write_queue))) {
+        if (c)
+            ret = write(c->fd, buffer_data(write_queue), buf_datacount);
+        else
+            ret = vchan_write_nonblock(d->vchan, buffer_data(write_queue), buf_datacount);
+        if (ret < 0) {
+            if (errno == EAGAIN || errno == EWOULDBLOCK) {
+                /* receiver doesn't have space for more data,
+                 * buffer actual requested data and exit */
+                buffer_append(write_queue, data, data_len);
+                return 1;
+            }
+            perror("vchan/client write");
+            return 0;
+        }
+        buffer_substract(write_queue, ret);
+    }
+
+    count = 0;
+    while (count < data_len) {
+        if (c)
+            ret = write(c->fd, data+count, data_len-count);
+        else
+            ret = vchan_write_nonblock(d->vchan, data+count, data_len-count);
+        if (ret < 0) {
+            if (errno == EAGAIN || errno == EWOULDBLOCK) {
+                /* receiver doesn't have space for more data,
+                 * buffer remaining requested data and exit */
+                buffer_append(write_queue, data+count, data_len-count);
+                return 1;
+            }
+            perror("vchan/client write");
+            return 0;
+        }
+        count += ret;
+    }
+    return 1;
 }
 
 int read_vchan_or_client(struct db_daemon_data *d, struct client *c,
@@ -175,8 +222,10 @@ int read_vchan_or_client(struct db_daemon_data *d, struct client *c,
             /* if vchan not connected, return error */
             return 0;
         if (libvchan_recv(d->vchan, data, data_len) < 0) {
-            perror("vchan read");
-            exit(1);
+            perror("closing vchan. vchan read");
+            libvchan_close(d->vchan);
+            d->vchan = NULL;
+            return 0;
         }
         return 1;
     } else {
@@ -608,21 +657,41 @@ int handle_vchan_multiread_resp(struct db_daemon_data *d, struct qdb_hdr *hdr) {
  * function. Any error in processing vchan data should be considered fatal.
  * @param d Daemon global data
  * @return 1 on success (message handled and responded), 0 if error
- *           occured and client should be disconnected.
+ *           occured and client should be disconnected, 2 if message not yet
+ *           handled (waiting for more data)
  */
 int handle_vchan_data(struct db_daemon_data *d) {
     struct qdb_hdr untrusted_hdr;
     struct qdb_hdr hdr;
 
-    if (libvchan_recv(d->vchan, &untrusted_hdr, sizeof(untrusted_hdr)) < 0) {
-        perror("vchan read");
-        return 0;
+    if (d->vchan_pending_hdr.type == QDB_INVALID_CMD) {
+        /* no previous header, retrieve it */
+        if (libvchan_data_ready(d->vchan) < sizeof(untrusted_hdr)) {
+            /* not enough data in vchan, wait for more */
+            return 2;
+        }
+        if (libvchan_recv(d->vchan, &untrusted_hdr, sizeof(untrusted_hdr)) < 0) {
+            perror("vchan read");
+            return 0;
+        }
+        if (!verify_hdr(&untrusted_hdr, 1)) {
+            fprintf(stderr, "invalid message received from peer\n");
+            return 0;
+        }
+        hdr = untrusted_hdr;
+    } else {
+        hdr = d->vchan_pending_hdr;
+        d->vchan_pending_hdr.type = QDB_INVALID_CMD;
     }
-    if (!verify_hdr(&untrusted_hdr, 1)) {
-        fprintf(stderr, "invalid message received from peer\n");
-        return 0;
+
+    /* This check is correct only because the whole message (up to
+     * QDB_MAX_DATA) can fit into a vchan buffer. Otherwise it could cause a
+     * deadlock. */
+    if (libvchan_data_ready(d->vchan) < hdr.data_len) {
+        /* save the header, but process the message when the rest is available */
+        d->vchan_pending_hdr = hdr;
+        return 2;
     }
-    hdr = untrusted_hdr;
 
     switch (hdr.type) {
         case QDB_CMD_WRITE:
@@ -630,6 +699,12 @@ int handle_vchan_data(struct db_daemon_data *d) {
                 return 0;
             break;
         case QDB_CMD_MULTIREAD:
+            if (d->remote_connected) {
+                fprintf(stderr, "received spurious QDB_CMD_MULTIREAD on vchan: "
+                                "only one after connection allowed\n");
+                discard_data_and_send_error(d, NULL, &hdr);
+                return 0;
+            }
             if (!handle_multiread(d, NULL, &hdr))
                 return 0;
             /* remote have synchronized database, send furher updates */
@@ -637,6 +712,13 @@ int handle_vchan_data(struct db_daemon_data *d) {
             break;
 
         case QDB_CMD_RM:
+            /* if there is no space for response, drop the command - remote
+             * side seems unresponsive */
+            if (libvchan_buffer_space(d->vchan) < sizeof(hdr)) {
+                fprintf(stderr, "got QDB_CMD_RM from remote domain, "
+                                "but there is no space in vchan for the reponse; dropping\n");
+                return 0;
+            }
             if (!handle_rm(d, NULL, &hdr))
                 return 0;
             break;
